/***********************************************

        COMPILING SPREADSHEET SOURCE CODE

        by Mike Morrow

        Dependencies (abbreviated):
                comp.exe : comp.c lex.c expr.c code.c comp_uti.c invoke.asm

        Header files:
                sys.h lex.h struct.h io.h

************************************************/




#include "sys.h"                        /* system wide definitions */
#include "lex.h"                        /* lexical analyzer defs   */
#include "struct.h"                     /* spreadsheet structure   */
#include "io.h"                         /* I/O parameters          */

#define DEBUG FALSE

#define NAN     (1.0/0.0)               /* value for "Not A Number" */

#define PRES_LOC curr_row][curr_col

#define UP -72
#define DOWN -80
#define LEFT -75
#define RIGHT -77

typedef int jmp_buf[3];
jmp_buf env;

extern token_type token;

/* The spreadsheet proper -- a 2D array of pointers to cell structures */
cell_ptr spread[NUM_ROW][NUM_COL];

static cell_ptr last;           /* the cell that points to the end of the
                                   program defined by the spreadsheet.     */

                                /* max_row & max_col give the lower right hand
                                   corner of a bounding box that encloses all */
static unsigned int max_row;   /* of the active cells.                     */
static char max_col;

main()
{
     void invoke();
     cell_ptr cellalloc();
     token_type yylex();

     cell_ptr curr;
     unsigned int curr_row;
     char curr_col;
     int action, moved;


/* INITIALIZE */
     init_spread(spread);
     init_disp();

     max_col = curr_col = 0;
     max_row = curr_row = 0;
     last = spread[PRES_LOC] = cellalloc();

     init_code();
     code_ret();
     last->jump_instr_loc = new_code(&(last->code));

/* SET UP SPREADSHEET DISPLAY */
     clrscr();
     show_cols(FIRST_COL);
     show_rows(FIRST_ROW);
     show_cell(spread[PRES_LOC], curr_row, curr_col);

/* MAIN LOOP */
     do{
                setjmp(env);
                moved = FALSE;

        /* GET INPUT FROM USER */
                erase_user_input();
                set_curs(curr_row, curr_col);
                action = user_input();

        /* DECIDE WHAT TO DO WITH USER REQUEST */
                switch (action)
                {
                        case RECALC:
                             start_exec((spread[0][0])->code);
                             show_all(spread, max_row, max_col);
                             break;

                        case EXPR:
                             yylex();
                             if(token == NUMBER || token == IDENT_TOK
                              ||token == '('    || token == '-'
                              || token == '+')
                             {
                                if(! spread[PRES_LOC])
                                            spread[PRES_LOC] =cellalloc();
                                compile(spread[PRES_LOC], curr_row, curr_col);
                             }

                             break;

                        case LABEL:
                             if(! spread[PRES_LOC])
                                  spread[PRES_LOC] = cellalloc();
                             label_cell(spread[PRES_LOC], curr_row, curr_col);

                             break;

                        case UP:
                             if(curr_row != 0)
                             {
                                curr_row--;
                                moved = TRUE;
                             }
                             break;

                        case DOWN:
                             if(curr_row < NUM_ROW - 1)
                             {
                                curr_row++;
                                moved = TRUE;
                             }
                             break;

                        case LEFT:
                             if(curr_col != 0)
                             {
                                curr_col--;
                                moved = TRUE;
                                }
                             break;

                        case RIGHT:
                             if(curr_col < NUM_COL - 1)
                             {
                                curr_col++;
                                moved = TRUE;
                             }
                             break;
                }
                if(moved)
                   show_cell(spread[PRES_LOC], curr_row, curr_col);

       }while(token != QUIT_TOK);
       clrscr();
}


/*
        init_spr(spr)           --       initialize spreadsheet "spr"
*/

void init_spread(spr)
cell_ptr spr[][NUM_COL];
{
        unsigned int i, j;


/*  No cells have been used yet -- set to NULL */
        for(i = 0; i<NUM_ROW; i++)
              for(j = 0; j<NUM_COL; j++) spr[i][j] = NULL;
}

/*
        compile(curr, row, col) --      transform user source input to a
                                        mach. language program.  This cell
                                        is bound to the program segment and
                                        to the user's source text.
*/
void compile(curr, row, col)
cell_ptr curr;
unsigned int row, col;
{
        init_code();
        arith_expr();                   /* parse & gen. code */
        if(! complete()) error("Extra chars at end of line");

        curr->type = NUMBER;
        code_assig(&(curr->data.value)); /* assignment to this cell */
        code_ret();                      /* code a return instruction */


        curr->jump_instr_loc = new_code(&(curr->code));

        start_exec(curr->code);         /* figure this cell's new value */


        new_text(&(curr->text));        /* link into spreadsheet */
        link_cell(curr, row, col);

        move_curs(row, col);            /* go to cell on monitor */
        show_val(curr->data.value);     /* show new value in cell */
        show_cell(curr, row, col);      /* show source text */
}


/*
        label_cell(curr, row, col) --   The indicated cell is to contain
                                        a label (text string).  The cell
                                        will contribute no code to the
                                        mach. language program being
                                        constructed.
*/

void label_cell(curr, row, col)
cell_ptr curr;
unsigned int row, col;
{
        curr->type = LABEL;

        init_code();
        code_ret();                     /* the program segment assoc'd with
                                           this cell does nothing      */

        curr->jump_instr_loc = new_code(&(curr->code));

        new_text(&(curr->text));
        curr->data.value = NAN;         /* default value for label cell */
        link_cell(curr, row, col);

        show_label(curr->text, row, col);
        show_cell(curr, row, col);
}



/*
        link_cell(curr, row, col) --    link the code bound to a cell into
                                        the rest of the code generated
                                        by the spreadsheet
*/
link_cell(curr, row, col)
cell_ptr curr;
unsigned int row, col;
{
        cell_ptr prev();

        cell_ptr previous;


        previous = prev(row, col);

        if(curr->next_code == NULL && curr != last) /* new cell */
        {
                if(previous)curr->next_code = previous->next_code;
                if(previous == last) last = curr;
        }

        if(curr->next_code)  /* should jump to next cell after done exec'ing */
                new_jump(curr->code, curr->jump_instr_loc, curr->next_code);

        if(previous)         /* if there is a prev cell, update it */
        {                    /* to jump to this cell next */
                previous->next_code = curr->code;
                new_jump(previous->code, previous->jump_instr_loc, curr->code);
        }

                /* new lower R.H. corner of bounding box? */
        if(max_row < row) max_row = row;
        if(max_col < col) max_col = col;

#if DEBUG
        dump_instrs(curr->code, curr->jump_instr_loc + 3);
#endif
}


/*
        prev(row, col) --      returns a pointer to the
                                        non-empty cell previous to the
                                        cell indicated by the args.
*/
cell_ptr prev(row, col)
unsigned int row, col;
{
        cell_ptr curr;
        unsigned int r, c;

        if(col == 0 && row == 0) return NULL;           /* (no prev. cell ) */

        r = row; c = col;
        do
        {
                if(c == 0)
                {
                        c = NUM_COL - 1;
                        r--;
                }
                else    c--;
                curr = spread[r][c];
        }while(! curr);
        return curr;
}

/*
        start_exec(loc) --      Execute machine code starting
                                at location "loc".
*/

void start_exec(loc)
unsigned int loc;
{
        void invoke();

        init_code();
        code_jmp(loc);
        invoke();
}


/*
        new_text(ptr)   --      put a new text string into the loc spec'd
                                by "ptr".  If a string was already at that
                                loc, free it.
*/

void new_text(ptr)
pointer *ptr;
{
        pointer zalloc();
        char *last_line();

        char *line_buffer;

        line_buffer = last_line();
        if(*ptr)free(*ptr);
        *ptr = zalloc(strlen(line_buffer) + 1);
        strcpy(*ptr, line_buffer);
}



/*
        cellalloc()             --      allocate memory for a new cell,
                                        initialize its fields.
*/
#define EMPTY_STR ""
cell_ptr cellalloc()
{
        cell_ptr zalloc();

        cell_ptr curr;

        curr = zalloc(sizeof(struct cell));
        curr->next_code = curr->code = NULL;
        curr->text = zalloc(strlen(EMPTY_STR) + 1);
        strcpy(curr->text, EMPTY_STR);
        curr->data.value = 0.0;
        curr->type = (char) NUMBER;
        curr->jump_instr_loc = 0;
        return curr;
}


/*
        *ret_val_ptr(c)

        transfer function -- return poitner to the value storage loc.
                             of the indicated cell.
*/

DATA_TYPE *ret_val_ptr(c)
cell_ptr c;
{
        return &(c->data.value);
}
